// 2D-KS $Id: Range Counting Tree
// Author:  Raul H. C. Lopes

/**********************************************************************************
 * Package: 2D-KS                                                                 *
 *                                                                                *
 * Description:                                                                   *
 *      Range Counting Tree                                                       *
 *                                                                                *
 * Author (alphabetical):                                                         *
 *      Raul H. Cardoso Lopes  - Brunel Unviversity                               *
 *                               Federal University of Espirito Santo             *
 *          http://www.inf.ufes.br/~/raulh                                        *
 *          email: rlopes at cern dot ch                                          *
 *                 rhclopes at yahoo dot com dot br                               *
 *                                                                                *
 * Copyright (c) 2007:                                                            *
 *      Brunel University                                                         *
 *                                                                                *
 * This program is free software; you can redistribute it and/or modify           *
 * it under the terms of the GNU General Public License as published by           *  
 * the Free Software Foundation; either version 2 of the License, or (at          *
 * your option) any later version.                                                * 
 *                                                                                *
 * This program is distributed in the hope that it will be useful, but            *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                     *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU              *
 * General Public License for more details.                                       *
 *                                                                                *
 *                                                                                *
 **********************************************************************************/


//# vi:sw=4 ts=4 sm et autoindent backspace=eol,start,indent lines=35 columns=120
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "conditions.h"
#include "ff.h"


#define max(x,y) (x<y ? y : x)

// 
// qdiff: 
//    p: pointer to a  point
//    rct0: range tree of first sample
//    rct1: range tree over second sample
//    RETURNS: maximum CDF difference in four quadrants centred in p
//
// Algorithm: computes 4 two-sided range queries in each range tree
//


double 
qdiff(point_t *r, rc_tree_t *rct0, rc_tree_t *rct1)
{
  coord_t xmax = max(rct0->xmax, rct1->xmax);
  coord_t ymax = max(rct0->ymax, rct1->ymax);
  double n0 = rct0->npoints;
  double n1 = rct1->npoints;

  //printf("ref pt: %lf %lf xmax %lf ymax %lf\n", r->x, r->y, xmax, ymax); 
  double ql0 = rct_query(rct0, r->x, ymax);
  double ql1 = rct_query(rct1, r->x, ymax);

  //printf("        ql0 %lf ql1 %lf\n", ql0,ql1);

  double qd0 = rct_query(rct0, xmax, r->y);
  double qd1 = rct_query(rct1, xmax, r->y);

  //printf("         qd0 %lf qd1 %lf\n", qd0, qd1);
  double q2rct0 = rct_query(rct0, r->x, r->y);
  double q2rct1 = rct_query(rct1, r->x, r->y);
  double q2diff = fabs(q2rct0/n0 - q2rct1/n1);

  //printf("new qdiff\n");
  //printf("        q2r0 %lf q2r1 %lf\n",  q2rct0, q2rct1);

  double q1rct0 = rct_query(rct0, r->x, ymax) - q2rct0;
  double q1rct1 = rct_query(rct1, r->x, ymax) - q2rct1;
  double q1diff = fabs(q1rct0/n0 - q1rct1/n1);

  //printf("        q1r0 %lf q1r1 %lf\n", q1rct0, q1rct1);

  double q3rct0 = rct_query(rct0, xmax, r->y) - q2rct0;
  double q3rct1 = rct_query(rct1, xmax, r->y) - q2rct1;
  double q3diff = fabs(q3rct0/n0 - q3rct1/n1);

  //printf("        q3r0 %lf q3r1 %lf\n", q3rct0, q3rct1);

  double q0rct0 = n0 - q1rct0 - q2rct0 - q3rct0;
  double q0rct1 = n1 - q1rct1 - q2rct1 - q3rct1;
  double q0diff = fabs(q0rct0/n0 - q0rct1/n1);

  //printf("        q0r0 %lf q0r1 %lf\n", q0rct0, q0rct1);

  //printf("rct0: %lf rct1: %lf\n", (q0rct0+q1rct0+q2rct0+q3rct0), (q0rct1+q1rct1+q2rct1+q3rct1));
  double max0 = max(q0diff, q1diff);
  double max1 = max(q2diff, q3diff);
  return max(max0,max1);
}

// statistic:
// ALGORITHM: 
//    first n0 points in pts are from first sample
//    next n1 points are from second sample
//
//    each point defines a quadrant
//    use qdiff to compute maximum CDF difference in each quadrant
//

double 
statistic(point_t *pts, int  n0, int n1)
{
  point_t *spts = (point_t *) malloc((n0+n1)*sizeof(point_t));
  assert_alloc(spts);
  memcpy(spts, pts, (n0+n1)*sizeof(point_t));

  rc_tree_t *rct0 = mk_rc_tree(pts, n0);
  rc_tree_t *rct1 = mk_rc_tree(pts+n0, n1);

  double maxDelta0 = 0.0;
  double maxDelta1 = 0.0;
 
  for (int i = 0; i<n0; ++i)
  {
    double max0 = qdiff(spts+i, rct0, rct1);
    maxDelta0 = max(maxDelta0, max0); 
  } 
  
  for (int i = 0, j = n0; i<n1; ++i, ++j)
  {
    double max1 = qdiff(spts+j, rct0, rct1);
    maxDelta1 = max(maxDelta1, max1); 
  } 
 return 0.5 * (maxDelta0 + maxDelta1);
}



//# vi:sw=2 ts=2 sm et cindent backspace=eol,start,indent lines=35 columns=120
