// 2D-KS $Id: Range Counting Tree
// Author:  Raul H. C. Lopes

/**********************************************************************************
 * Package: 2D-KS                                                                 *
 *                                                                                *
 * Description:                                                                   *
 *      Range Counting Tree                                                       *
 *                                                                                *
 * Author (alphabetical):                                                         *
 *      Raul H. Cardoso Lopes  - Brunel Unviversity                               *
 *                               Federal University of Espirito Santo             *
 *          http://www.inf.ufes.br/~/raulh                                        *
 *          email: rlopes at cern dot ch                                          *
 *                 rhclopes at yahoo dot com dot br                               *
 *                                                                                *
 * Copyright (c) 2007:                                                            *
 *      Brunel University                                                         *
 *                                                                                *
 * This program is free software; you can redistribute it and/or modify           *
 * it under the terms of the GNU General Public License as published by           *  
 * the Free Software Foundation; either version 2 of the License, or (at          *
 * your option) any later version.                                                * 
 *                                                                                *
 * This program is distributed in the hope that it will be useful, but            *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                     *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU              *
 * General Public License for more details.                                       *
 *                                                                                *
 *                                                                                *
 **********************************************************************************/


#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include"conditions.h"

#include "rct.h"


ysplit_t*
mk_ysplit(point_t *ptref, int lydisp, int rydisp, int lcount)
{
 ysplit_t * ysplt = (ysplit_t *) malloc (sizeof(ysplit_t));
 assert_alloc(ysplt);
 
 ysplt->ptref = ptref;
 ysplt->lydisp = lydisp;
 ysplt->rydisp = rydisp;
 ysplt->lcount = lcount;
 return ysplt;
}


void
print_inode(rct_inode_t *nd)
{

 if (!nd) return;
 printf("nd %p med %lf %lf left %p right %p\n", nd, nd->medref->x, nd->medref->y, nd->left, nd->right);
 for (int i=0; i < nd->nysplit; ++i)
 {
   ysplit_t * yvec = nd->yvec;
   printf("        %lf %lf %d %d %d\n", yvec[i].ptref->x, 
                                        yvec[i].ptref->y, yvec[i].lydisp, yvec[i].rydisp, yvec[i].lcount);
 }
}


rct_inode_t *
mk_rct_inode(point_t *medref, rct_inode_t *left, rct_inode_t *right, ysplit_t *yvec, int nysplit)
{
  rct_inode_t *nd = (rct_inode_t *) malloc(sizeof(rct_inode_t));
  assert_alloc(nd);

  nd->medref = medref;
  nd->left = left;
  nd->right = right;
  nd->yvec = yvec;
  nd->nysplit = nysplit;
  //print_inode(nd);
  //print_inode(left);
  //print_inode(right);
  return nd;
}

#define isLeaf(nd) (nd->left == 0)

int
ysplit_cmp(const void *c0, const void*c1)
{
  ysplit_t *ysplt0 = (ysplit_t *) c0;
  ysplit_t *ysplt1 = (ysplit_t *) c1;
 
  if (ysplt0->ptref->y < ysplt1->ptref->y) return -1;
  if (ysplt0->ptref->y > ysplt1->ptref->y) return 1;
  return 0;
}




ysplit_t *
yvec_merge(ysplit_t *lvec, int nlvec, ysplit_t *rvec, int nrvec)
{
  int nmvec = nlvec + nrvec;
  ysplit_t *mvec = (ysplit_t *) malloc(nmvec*sizeof(ysplit_t));
  assert_alloc(mvec);

  memcpy(mvec, lvec, sizeof(ysplit_t) * nlvec);
  for (int i = 0; i < nlvec; ++i)
    mvec[i].lcount = 1;
  memcpy(mvec+nlvec, rvec, sizeof(ysplit_t) * nrvec);
  for (int i = nlvec; i < nlvec+nrvec; ++i)
    mvec[i].lcount = 0;

  qsort(mvec, nmvec, sizeof(ysplit_t), ysplit_cmp);

  for (int i=1, j=0; i < nmvec; ++i, ++j)
  {
    // prefix sum
    mvec[i].lcount += mvec[j].lcount;

    // bcast
    if (mvec[i].lydisp < 0)
      mvec[i].lydisp = mvec[j].lydisp;

    // bcast
    if (mvec[i].rydisp < 0)
      mvec[i].rydisp = mvec[j].rydisp;
  }

  return mvec;
}

int 
point_cmp(const void *c0, const void *c1)
{
  point_t *p0, *p1;

  p0 = (point_t *) c0;
  p1 = (point_t *) c1;
  if (p0->x < p1->x) return -1;
  if (p0->x > p1->x) return 1;
  return 0;
}

rct_inode_t *
mk_rct_leaf(point_t *p)
{
  return mk_rct_inode(p, 0, 0, 0, 0);
}


rct_inode_t *
mk_index(point_t *pts, int lb, int n)
{
  assert(n!=0);
  if (1 == n)
  {
    ysplit_t *ysplt = (ysplit_t *) malloc(sizeof(ysplit_t));
    assert_alloc(ysplt);
    ysplt->ptref = pts+lb;
    ysplt->lydisp = 0;
    ysplt->rydisp = 0;
    ysplt->lcount = 1;
    return mk_rct_inode(pts+lb, mk_rct_leaf(pts+lb), 0, ysplt, 1);
  }
  if (2 == n)
  {
    ysplit_t *ysplt = (ysplit_t *) malloc(2*sizeof(ysplit_t));
    assert_alloc(ysplt);
    if (pts[lb].y <= pts[lb+1].y)
    {
      ysplt[0].ptref = pts+lb;
      ysplt[0].lydisp = 0; 
      ysplt[0].rydisp = 0;
      ysplt[0].lcount = 1;

      ysplt[1].ptref = pts+lb+1;
      ysplt[1].lydisp = 0; 
      ysplt[1].rydisp = 0;
      ysplt[1].lcount = 1;
      return mk_rct_inode(pts+lb, mk_rct_leaf(pts+lb), mk_rct_leaf(pts+lb+1), ysplt, 2);
    }
    else
    {
      ysplt[0].ptref = pts+lb+1;
      ysplt[0].lydisp = 0; 
      ysplt[0].rydisp = 0;
      ysplt[0].lcount = 0;

      ysplt[1].ptref = pts+lb;
      ysplt[1].lydisp = 0; 
      ysplt[1].rydisp = 0;
      ysplt[1].lcount = 1;
      return mk_rct_inode(pts+lb, mk_rct_leaf(pts+lb), mk_rct_leaf(pts+lb+1), ysplt, 2);
    }
    
  }
  int nleft = (n % 2 ? (n>>1)+1 : n>>1) ;
  rct_inode_t *linode = mk_index(pts, lb, nleft);
  rct_inode_t *rinode = mk_index(pts, lb+nleft, n-nleft);
  ysplit_t * mvec = yvec_merge(linode->yvec, linode->nysplit, rinode->yvec, rinode->nysplit); 
  return mk_rct_inode(pts+lb+nleft-1, linode, rinode, mvec, n);
}

rc_tree_t *
mk_rc_tree(point_t *pts, int n)
{
  qsort(pts, n, sizeof(point_t), point_cmp);
  rct_inode_t * ridx = mk_index(pts, 0, n);
  rc_tree_t * rct = (rc_tree_t *) malloc(sizeof(rc_tree_t));
  rct->points = pts;
  rct->npoints = n;
  rct->tree = ridx;
  rct->xmax = pts[n-1].x;
  rct->ymax = ridx->yvec[n-1].ptref->y;
  return rct;
}

int
yvec_search(ysplit_t *yvec, int nysplit, coord_t yt)
{
  int lb = 0,
      ub = nysplit;
  while (1 != (ub - lb))
  {
    int avg = (ub + lb)/2;
    if ( yt < yvec[avg].ptref->y)
      ub = avg;
    else 
      lb = avg;
  }
  return lb;
}

#define min(x,y) (x <= y ? x : y)
#define yvec_lcount(yvec,ydisp) yvec[ydisp].lcount

int 
rct_query(rc_tree_t * rct, coord_t xr, coord_t yt)
{
  rct_inode_t *r = rct->tree;
  int ydisp = r->nysplit; //yvec_search(r->yvec, r->nysplit, yt);
  int lcount = 0;
  while (!isLeaf(r))
  {
    ysplit_t * yvec = r->yvec;
    ydisp = yvec_search(yvec, min(ydisp+1, r->nysplit), yt);
    if (xr > inode_median(r)->x)
    {
      lcount += yvec_lcount(yvec, ydisp);
      r = r->right;
      if (!r) return lcount;
    }
    else
    {
      r = r->left;
    }
  }
  if (xr >= inode_median(r)->x && yt >= inode_median(r)->y)
    return (lcount+1);
  else 
    return lcount;
}


 
//# vi:sw=2 ts=2 sm et cindent backspace=eol,start,indent lines=35 columns=120 
